import numpy as np
from skretrieval.core.radianceformat import RadianceGridded
import sasktran as sk
from skretrieval.core.sensor.spectrograph import SpectrographFast
from skretrieval.core.lineshape import Gaussian, DeltaFunction
from sklearn import linear_model, metrics
from scipy import optimize, signal
from skretrieval.util import Timer


def o3_xs(mean_lat, mean_lon, mean_alt, mean_mjd, wavel_grid):
    o3_optprop = sk.O3DBM()

    hires_o3 = o3_optprop.calculate_cross_sections(sk.MSIS90(), mean_lat, mean_lon, mean_alt, mean_mjd,
                                                   wavel_grid).total

    # Normalize each to mean 0, stdev 1
    hires_o3 -= np.nanmean(hires_o3)
    hires_o3 /= np.nanstd(hires_o3)
    return {'x': hires_o3, 'resolution': 0.02}

def no2_xs(mean_lat, mean_lon, mean_alt, mean_mjd, wavel_grid):
    no2_optprop = sk.NO2Vandaele1998()

    hires_no2 = no2_optprop.calculate_cross_sections(sk.MSIS90(), mean_lat, mean_lon, mean_alt, mean_mjd,
                                                     wavel_grid).total

    # Normalize each to mean 0, stdev 1
    hires_no2 -= np.nanmean(hires_no2)
    hires_no2 /= np.nanstd(hires_no2)
    return {'x': hires_no2, 'resolution': np.nanmean(no2_optprop.info['spectral resolution'])}

def o2o2_xs(mean_lat, mean_lon, mean_alt, mean_mjd, wavel_grid):
    o2o2_optprop = sk.O2O2HITRAN2016()

    hires_o2o2 = o2o2_optprop.calculate_cross_sections(sk.MSIS90(), mean_lat, mean_lon, mean_alt, mean_mjd,
                                                       wavel_grid).total
    # Normalize each to mean 0, stdev 1
    hires_o2o2 -= np.nanmean(hires_o2o2)
    hires_o2o2 /= np.nanstd(hires_o2o2)
    return {'x': hires_o2o2, 'resolution': 0}

def solar_spectrum(mean_mjd, wavel_grid):
    solar_spectrum = sk.SolarSpectrum()

    hires_solar = solar_spectrum.irradiance(wavel_grid, mjd=mean_mjd)

    # Normalize each to mean 0, stdev 1
    hires_solar -= np.nanmean(hires_solar)
    hires_solar /= np.nanstd(hires_solar)

    return {'x': hires_solar, 'resolution': 0.04}

def convolve(hires, hires_wavel, lowres_wavel, fwhm, shift):
    shape = Gaussian(fwhm=fwhm, mode='linear')

    weights = shape.integration_weights(shift, hires_wavel - np.nanmean(hires_wavel))
    hires_c = signal.fftconvolve(hires, weights, mode='same')
    lowres = np.interp(lowres_wavel, hires_wavel, hires_c)

    return lowres


def spectral_fit(l1_data: RadianceGridded, fit_alt_range=(30000, 34000),
                 fit_spectral_range=(435, 475), high_alt_normalize=False,
                 high_alt_normalize_range=(32000, 34000),
                 include_ozone=True,
                 include_no2=True,
                 include_o2o2=False):
    """"""
    hires_wavel_grid = np.linspace(fit_spectral_range[0] - 5, fit_spectral_range[1] + 5, 10000)

    results = dict()

    slit_data = l1_data.data.sel(wavelength=slice(fit_spectral_range[0], fit_spectral_range[1]))
    tangent_locations = l1_data.tangent_locations()

    if high_alt_normalize:
        slit_data = slit_data.where((tangent_locations.altitude > fit_alt_range[0]) & (tangent_locations.altitude < fit_alt_range[1])).mean(dim='los') / \
                    slit_data.where((tangent_locations.altitude > high_alt_normalize_range[0]) & (
                                tangent_locations.altitude < high_alt_normalize_range[1])).mean(dim='los')
    else:
        slit_data = slit_data.where((tangent_locations.altitude > fit_alt_range[0]) & (tangent_locations.altitude < fit_alt_range[1])).mean(dim='los')

    mean_tangent = tangent_locations.where((tangent_locations.altitude > fit_alt_range[0]) & (tangent_locations.altitude < fit_alt_range[1])).mean(dim='los')

    mean_lat, mean_lon, mean_alt = float(mean_tangent.latitude), float(mean_tangent.longitude), float(mean_tangent.altitude)
    mean_mjd = float(slit_data.mjd)

    predictors = []
    if include_ozone:
        predictors.append(o3_xs(mean_lat, mean_lon, mean_alt, mean_mjd, hires_wavel_grid))
    if include_no2:
        predictors.append(no2_xs(mean_lat, mean_lon, mean_alt, mean_mjd, hires_wavel_grid))
    if include_o2o2:
        predictors.append(o2o2_xs(mean_lat, mean_lon, mean_alt, mean_mjd, hires_wavel_grid))

    if not high_alt_normalize:
        predictors.append(solar_spectrum(mean_mjd, hires_wavel_grid))

    reg_model = linear_model.LinearRegression()
    fit_data = np.log(slit_data.radiance.values)
    fit_data -= np.nanmean(fit_data)
    fit_data /= np.nanstd(fit_data)

    non_nan = ~np.isnan(fit_data)

    fit_data = fit_data[non_nan] - np.polyval(
        np.polyfit(slit_data.wavelength.values[non_nan], fit_data[non_nan], deg=3),
        slit_data.wavelength.values[non_nan])

    def fit_fn(x):
        shift = x[0]
        fwhm = x[1]
        lowres_pred = []
        for pred in predictors:
            lowres_fwhm = np.sqrt(fwhm**2 - pred['resolution']**2)
            lowres_pred.append(convolve(pred['x'], hires_wavel_grid, slit_data.wavelength.values, lowres_fwhm, shift))

            lowres_pred[-1] = lowres_pred[-1] - np.polyval(
                np.polyfit(slit_data.wavelength.values, lowres_pred[-1], deg=3),
                slit_data.wavelength.values)

        lowres_pred = np.array(lowres_pred).T

        reg_model.fit(lowres_pred[non_nan], fit_data)

        predictions = reg_model.predict(lowres_pred[non_nan])

        return predictions - fit_data

    # return metrics.median_absolute_error(fit_data, predictions)

    # result = optimize.minimize(fit_fn, x0=(-0.2, 0.3), bounds=((-0.5, 0.5), (0.1, 0.7)))
    result = optimize.least_squares(fit_fn, x0=(0, 1.25), loss='cauchy', bounds=((-0.5, 0.1), (0.5, 1.5)))

    results['shift'] = result['x'][0]
    results['fwhm'] = result['x'][1]
    results['raw'] = result

    shift = result['x'][0]
    fwhm = result['x'][1]
    lowres_pred = []
    for pred in predictors:
        lowres_fwhm = np.sqrt(fwhm ** 2 - pred['resolution'] ** 2)
        lowres_pred.append(convolve(pred['x'], hires_wavel_grid, slit_data.wavelength.values, lowres_fwhm, shift))

        lowres_pred[-1] = lowres_pred[-1] - np.polyval(
            np.polyfit(slit_data.wavelength.values, lowres_pred[-1], deg=3),
            slit_data.wavelength.values)

    lowres_pred = np.array(lowres_pred).T

    reg_model.fit(lowres_pred[non_nan], fit_data)

    predictions = reg_model.predict(lowres_pred[non_nan])

    shift = 0
    fwhm = 1.25
    lowres_pred = []
    for pred in predictors:
        lowres_fwhm = np.sqrt(fwhm ** 2 - pred['resolution'] ** 2)
        lowres_pred.append(convolve(pred['x'], hires_wavel_grid, slit_data.wavelength.values, lowres_fwhm, shift))

        lowres_pred[-1] = lowres_pred[-1] - np.polyval(
            np.polyfit(slit_data.wavelength.values, lowres_pred[-1], deg=3),
            slit_data.wavelength.values)

    lowres_pred = np.array(lowres_pred).T

    reg_model.fit(lowres_pred[non_nan], fit_data)

    predictions_before = reg_model.predict(lowres_pred[non_nan])

    results['radiance'] = fit_data
    results['wavel'] = slit_data.wavelength.values[non_nan]
    results['radiance_fit'] = predictions
    results['radiance_before_fit'] = predictions_before


    return results
