from osiris_skretrieval.osiris_target import OsirisTarget
from skretrieval.core.radianceformat import RadianceBase
import numpy as np
import xarray as xr
from scipy import interpolate


class OsirisTriplet(OsirisTarget):
    def __init__(self, triplets, combine_y=False, high_altitude_normalize_jacobian=False,
                 use_full_normalization_range=True, linear_combine_triplets=False, **kwargs):
        self._triplets = triplets
        self._combine_y = combine_y
        self._high_altitude_normalize = True
        self._high_altitude_normalize_jacobian = high_altitude_normalize_jacobian
        self._use_full_normalization_range = use_full_normalization_range
        self._linear_combine_triplets = linear_combine_triplets
        super().__init__(**kwargs)

    def measurement_wavelengths(self, measurement_l1: RadianceBase):
        wavel = []
        for triplet in self._triplets:
            triplet_wavel = measurement_l1.data.sel(wavelength=triplet['wavelengths'], method='nearest').wavelength.values
            wavel.append(triplet_wavel)
        return np.unique(np.hstack(wavel))

    def measurement_vector(self, l1_data: RadianceBase):

        tangent_locations = l1_data.tangent_locations()
        good_data = l1_data.data.copy(deep=True)
        # Filter out below the lowest altitude
        good_data['tangent_altitude'] = tangent_locations['altitude']
        good_data = good_data.swap_dims({'los': 'tangent_altitude'})
        good_data = good_data.where((good_data.tangent_altitude > self._retrieval_altitudes[0] - (self._retrieval_altitudes[1] - self._retrieval_altitudes[0]) / 2), drop=True)

        all_y = []
        all_wf = []
        all_dy = []
        all_y_altitudes = []
        for triplet in self._triplets:
            triplet_data = good_data.sel(wavelength=triplet['wavelengths'], method='nearest')

            log_triplet_data = np.log(triplet_data['radiance'])

            if not self._use_full_normalization_range:
                # ignore the normalization range and use the first altitude above the low normalization altitude that
                # isn't nan
                normalization_data = triplet_data.where(
                    (triplet_data.tangent_altitude > triplet['normalization_range'][0]), drop=True)
                normalization_radiance = normalization_data['radiance']

                temp_radiance = np.log(normalization_radiance.isel(tangent_altitude=0))
                c = 1
                while (np.any(np.isnan(temp_radiance))):
                    temp_radiance = np.log(normalization_radiance.isel(tangent_altitude=c))
                    c += 1

                normalization_radiance = temp_radiance
            else:
                # use the full normalization range, ignoring nans
                normalization_data = triplet_data.where(
                    (triplet_data.tangent_altitude > triplet['normalization_range'][0]) & (
                                triplet_data.tangent_altitude < triplet['normalization_range'][1]), drop=True)
                normalization_radiance = normalization_data['radiance']

                normalization_radiance = np.log(normalization_radiance.mean(dim='tangent_altitude'))

            log_triplet_data = log_triplet_data.where((log_triplet_data.tangent_altitude > triplet['lower_altitude_m']) & (log_triplet_data.tangent_altitude < triplet['upper_altitude_m']), drop=True)

            if self._high_altitude_normalize:
                normalized_log_triplet_data = log_triplet_data - normalization_radiance
            else:
                normalized_log_triplet_data = log_triplet_data

            # Check to see if our weights are a function of altitude
            if len(np.array(triplet['weights']).shape) == 1:
                triplet_weights = xr.DataArray(triplet['weights'],
                                               coords=[normalized_log_triplet_data.wavelength],
                                               dims=['wavelength'])
            else:
                # Interpolate the weights onto the tangent altitudes
                weights = interpolate.interp1d(triplet['weight_altitudes'], triplet['weights'], axis=1)(normalized_log_triplet_data.tangent_altitude.values)
                triplet_weights = xr.DataArray(weights,
                                               coords=[normalized_log_triplet_data.wavelength, normalized_log_triplet_data.tangent_altitude],
                                               dims=['wavelength', 'tangent_altitude'])

            y_triplet = normalized_log_triplet_data.dot(triplet_weights, dims='wavelength')

            if 'wf' in good_data:
                wf_data = (triplet_data['wf'] / triplet_data['radiance']).to_dataset(name='wf')
                wf_data = wf_data.where(
                    (wf_data.tangent_altitude > triplet['lower_altitude_m']) & (
                                wf_data.tangent_altitude < triplet['upper_altitude_m']), drop=True)

                if self._high_altitude_normalize_jacobian and self._high_altitude_normalize:
                    wf_data -= (normalization_data['wf'] / normalization_data['radiance']).mean(dim='tangent_altitude')

                if self._state_vector_space == 'log':
                    wf_data['x'] = (['perturbation'], self.state_vector())
                    wf_data['wf'] *= np.exp(wf_data['x'])

                wf = wf_data['wf'].dot(triplet_weights, dims='wavelength')
                all_wf.append(wf.values)

            if 'radiance_noise' in good_data:
                inverse_snr = (good_data['radiance_noise'] / good_data['radiance']).sel(wavelength=triplet['wavelengths'], method='nearest')

                inverse_snr = inverse_snr.where(
                    (inverse_snr.tangent_altitude > triplet['lower_altitude_m']) & (
                                inverse_snr.tangent_altitude < triplet['upper_altitude_m']), drop=True)

                dy = (inverse_snr**2).dot(triplet_weights**2, dims='wavelength')
                all_dy.append(dy.values)


            all_y.append(y_triplet.values)
            all_y_altitudes.append(normalized_log_triplet_data.tangent_altitude.values)
        y_out = dict()

        if not self._linear_combine_triplets:
            y_out['y'] = np.hstack(all_y)
            y_out['y_altitude'] = np.hstack(all_y_altitudes)

            y_out['triplet_y'] = all_y
            y_out['triplet_y_altitudes'] = all_y_altitudes

            if 'wf' in good_data:
                y_out['jacobian'] = np.vstack(all_wf)

            if 'radiance_noise' in good_data:
                y_out['y_error'] = np.hstack(all_dy).flatten()
        else:
            lowest_altitude = np.nanmin([triplet['lower_altitude_m'] for triplet in self._triplets])
            highest_altitude = np.nanmax([triplet['upper_altitude_m'] for triplet in self._triplets])

            useable_tangent_altitudes = tangent_locations.altitude.values
            useable_tangent_altitudes = useable_tangent_altitudes[(useable_tangent_altitudes > lowest_altitude) & (useable_tangent_altitudes < highest_altitude)
                                                                  & (useable_tangent_altitudes > self._retrieval_altitudes[0] - self._ret_spacing_m/2.0)]

            y_out['y'] = np.zeros((len(useable_tangent_altitudes)))
            y_out['y_altitude'] = useable_tangent_altitudes

            all_triplet_coef = []
            for triplet, y, y_altitude in zip(self._triplets, all_y, all_y_altitudes):
                triplet_coef = np.interp(useable_tangent_altitudes, triplet['wik_altitudes'], triplet['wik_values'], left=triplet['wik_values'][0], right=triplet['wik_values'][-1])
                y_add = np.interp(useable_tangent_altitudes, y_altitude, y, left=0, right=0)
                nan_y = np.isnan(y_add)
                triplet_coef[nan_y] = 0.0
                all_triplet_coef.append(triplet_coef)

            for triplet, y, y_altitude, triplet_coef in zip(self._triplets, all_y, all_y_altitudes, all_triplet_coef):
                y_add = np.interp(useable_tangent_altitudes, y_altitude, y, left=0, right=0)
                # Don't add nans, just set them to 0
                y_add[np.isnan(y_add)] = 0.0
                y_out['y'] += triplet_coef * y_add

            if 'radiance_noise' in good_data:
                y_out['y_error'] = np.zeros((len(useable_tangent_altitudes)))
                for triplet, y, y_altitude, y_error, triplet_coef in zip(self._triplets, all_y, all_y_altitudes, all_dy, all_triplet_coef):
                    dy_add = np.interp(useable_tangent_altitudes, y_altitude, y_error, left=0, right=0)
                    dy_add[np.isnan(dy_add)] = 0
                    y_out['y_error'] += dy_add * triplet_coef**2
                y_out['y_error'][y_out['y_error'] == 0] = 10000

            if 'wf' in good_data:
                y_out['jacobian'] = np.zeros((len(useable_tangent_altitudes), all_wf[0].shape[1]))
                for triplet, y, y_altitude, wf, triplet_coef in zip(self._triplets, all_y, all_y_altitudes, all_wf, all_triplet_coef):
                    wf_add = interpolate.interp1d(y_altitude, wf, axis=0, bounds_error=False, fill_value=0)(useable_tangent_altitudes)
                    wf_add[np.isnan(wf_add)] = 0.0
                    y_out['jacobian'] += wf_add * triplet_coef[:, np.newaxis]

        return y_out